/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef CONTAINER_PSD_HH
#define CONTAINER_PSD_HH

#include "fSeries.hh"

namespace containers {

    class DFT;

    /**  The PSD class is based on the fSeries class and contains a 
      *  power-spectral-density. THe data are normalized as specified in 
      *  "Conventions for data and software products of LIGO and the LSC" 
      *  (<a href="http://www.ligo.caltech.edu/docs/T/T010095-00.pdf">
      *  LIGO-T010095-00-Z</a>).
      *  @brief Power Spectral Density container class.
      *  @author John G. Zweizig
      *  @version 1.0; Last modified December 4, 2007
      */
    class PSD : public fSeries {
    public:
	/**  The default PSD constructor creates an empty PSD object
	  *  with no allocated data vector and invalid frequency step.
	  *  @brief Default PSD constructor.
	  */
	PSD(void);

	/**  Construct a PSD from the square of a DFT. The
	  *  fSeries data are filled as described under SetData.
	  *  @brief Construct a PSD from a DFT.
	  *  @param fs DFT from which the PSD is derived.
	  */
	PSD(const DFT& fs);


	/**  Destroy a PSD and release any allocated storage.
	  *  @brief Destroy a PSD.
	  */
	~PSD(void);

	/**  Create an identical copy of a PSD object.
	  *  @brief Clone a PSD
	  *  @return A pointer to the new PSD object.
	  */
	PSD* clone(void) const;

	/**  Get the DFT series type (kPSD).
	  *  @return The DFT type code (fSeries::kPSD)
	  */
	fSeries::FSType getFSType(void) const;

	/**  Fold the PSD at <tt>f=0</tt>.
	  */
	void fold(void);
    };

    //==================================  Inline methods
    inline fSeries::FSType 
    PSD::getFSType(void) const {
	return fSeries::kPSD;
    }

} // namespace containers

#endif // !defined(CONTAINER_PSD_HH)
