#include <string>
#include "TSeries.hh"
#include "Time.hh"
#include "Interval.hh"
#include "Histogram1.hh"
 
// a class to keep channel status info.

// fSumCount = number of main loops during fLogInterval
class ChannelStatus
{

public:
  enum {
    // normal
    kNone = 0, 
    // during event
    kEvent = 1, 
    // waiting event end
    kWait = 2

  };

  enum { // filter mode
    // no filter
    kNoFilter = -1,
    // low pass
    kLowPass = 0,
    // high pass
    kHighPass = 1,
    // band pass
    kBandPass = 2
  };

  TSeries               *fTSold,*fTSnew;
  /// Recent band-limited data
  TSeries               *fTSfiltered;
  /// Current 'Log Interval' histogram
  Histogram1            *fHistCur;
  /// Total events histogram
  Histogram1            *fHistTot;

  // channel name
  std::string fName;

  // sampling frequency
  int fSampleFreq;

  // number of CheckEvent executed during fLogInterval
  int fCheckEventCount;

  // number of events in fLogInterval
  int fEventCount;

  // 0:no event, 1:during event, 2:waititng end of event
  int fEventStatus;

  // number of data point in the event
  int fEventNPoint;

  // number of main loops executed for initializing statistics
  int fInitCount;

  // data value at maximum in the event
  float fMaxData;

  // sum of mean during fLogInterval
  float fSumMean;

  // mean value of time series
  float fMean;

  // mean value of time series at max in the event
  float fMaxMean;

  // normalization factor
  float fLastNorm;

  // standard deviation of time series
  float fSDev;

  // standard deviation of time series at max in the event
  float fMaxSDev;

  float fSumSDev;
  int fSumSDevCount;
  // lower threshold
  float fThresholdLow;

  // higher threshold (event above this level always triggered)
  float fThresholdHi;

  // cut off frequencies (lower and higher edges)
  float fCutOffFreqLow,fCutOffFreqHi;

  // filter order
  int fFilterOrder;

  // filter mode (0:lowpass 1:highpass 2:bandpass)
  int fFilterMode;

  // IIR filter coefficients (denominator & numerator) 
  double *fCoefA,*fCoefB;

  // start and end time of an evnet
  Time fStartTime,fEndTime;

  // time for max of the event
  Time fMaxTime;

  // trigger comment
  std::string fTrigComment;

  ChannelStatus() : fCoefA(0),fCoefB(0)
  {
    fTSnew = 0;
    fTSold = 0;
    fTSfiltered = 0;
    fHistCur = 0;
    fHistTot = 0;
    Reset();
    fInitCount = 1;
  }

  ChannelStatus(const ChannelStatus& ch) : fCoefA(0),fCoefB(0)
  {
    fTSnew = 0;
    fTSold = 0;
    fTSfiltered = 0;
    fHistCur = 0;
    fHistTot = 0;
    *this = ch; 
  }

  ~ChannelStatus(void)
  {
    if (fTSnew) delete fTSnew;
    if (fTSold) delete fTSold;
    if (fTSfiltered) delete fTSfiltered;
    if (fHistCur) delete fHistCur;
    if (fHistTot) delete fHistTot;
    if (fCoefA) delete[] fCoefA;
    if (fCoefB) delete[] fCoefB;
  }


  // clear all variables
  void Reset()
  {
    if (fTSnew) delete fTSnew;
    fTSnew = 0;
    if (fTSold) delete fTSold;
    fTSold = 0;
    if (fTSfiltered) delete fTSfiltered;
    fTSfiltered = 0;
    if (fHistCur) delete fHistCur;
    fHistCur = 0;
    // don't delete fHistTot

    fName = "";
    fSampleFreq = 0;
    fCheckEventCount = fEventCount = fEventNPoint = 0;
    fEventStatus = kNone;
    fInitCount = 0;
    fMaxData = 0;
    fSumMean = fMean = fMaxMean = 0;
    fSumSDev = 0;
    fSumSDevCount = 0;
    fLastNorm = fSDev = fMaxSDev = 0;
    fThresholdLow = fThresholdHi = 0;
    fCutOffFreqLow = fCutOffFreqHi = 0;
    fFilterOrder = 0;
    fFilterMode = kNoFilter;
    if (fCoefA) delete[] fCoefA;
    fCoefA = 0;
    if (fCoefB) delete[] fCoefB;
    fCoefB = 0;
    fStartTime = fEndTime = fMaxTime = 0;
    fTrigComment = "";
  }

  void ResetFilterCoef() 
  {
    int arraysize;
    switch (fFilterMode) {
    case kLowPass:
    case kHighPass:
      arraysize = fFilterOrder + 1;
      break;
    case kBandPass:
      arraysize = fFilterOrder * 2 + 1;
      break;
    default: // no filter
      arraysize = 0;
    }

    if (fCoefA) delete[] fCoefA;
    if (fFilterOrder) {
      fCoefA = new double[arraysize];
    }
    if (fCoefB) delete[] fCoefB;      
    if (fFilterOrder) {
      fCoefB = new double[arraysize];
    }

  }

  double GetEventDuration() const 
  {
    return fEndTime - fStartTime;
  }

  ChannelStatus& operator = (const ChannelStatus& ch)
  {

    Reset();

    if (ch.fTSnew)
      *fTSnew        = *ch.fTSnew;

    if (ch.fTSold)
      *fTSold        = *ch.fTSold;
    
    if (ch.fTSfiltered)
      *fTSfiltered   = *ch.fTSfiltered;
    if (ch.fHistCur)
      *fHistCur      = *ch.fHistCur;

    if (ch.fHistTot)
      *fHistTot      = *ch.fHistTot;

    fName            = ch.fName;
    fSampleFreq      = ch.fSampleFreq;
    fLastNorm        = ch.fLastNorm;
    fCheckEventCount = ch.fCheckEventCount;
    fEventCount      = ch.fEventCount;
    fEventNPoint     = ch.fEventNPoint;
    fEventStatus     = ch.fEventStatus;
    fInitCount       = ch.fInitCount;
    fMaxData         = ch.fMaxData;
    fSumMean         = ch.fSumMean;
    fMean            = ch.fMean;
    fSDev            = ch.fSDev;
    fSumSDev         = ch.fSumSDev;
    fSumSDevCount    = ch.fSumSDevCount;
    fMaxMean         = ch.fMaxMean;
    fLastNorm        = ch.fLastNorm;
    fMaxSDev         = ch.fMaxSDev;
    fThresholdLow    = ch.fThresholdLow;
    fThresholdHi     = ch.fThresholdHi;
    fCutOffFreqLow   = ch.fCutOffFreqLow;
    fCutOffFreqHi    = ch.fCutOffFreqHi;
    fStartTime       = ch.fStartTime;
    fEndTime         = ch.fEndTime;
    fMaxTime         = ch.fMaxTime;
    fTrigComment     = ch.fTrigComment;
    fFilterOrder     = ch.fFilterOrder;
    fFilterMode      = ch.fFilterMode;

    ResetFilterCoef();

    int arraysize;
    switch (ch.fFilterMode) {
    case kLowPass:
    case kHighPass:
      arraysize = ch.fFilterOrder + 1;
      break;
    case kBandPass:
      arraysize = ch.fFilterOrder * 2 + 1;
      break;
    default:
      arraysize = ch.fFilterOrder + 1;
    }

    if (ch.fCoefA) {
      fCoefA = new double[arraysize];
      memcpy(fCoefA,ch.fCoefA,arraysize);
    }

    if (ch.fCoefB) {
      fCoefB = new double[arraysize];
      memcpy(fCoefB,ch.fCoefB,arraysize);
    }

    return *this;
  }

  bool IsTSContinuous() 
  {
    if (fTSnew->getStartTime().getS() - 
      fTSold->getStartTime().getS() == 1 ) return true;
    else return false;
  }

  void CopyTStoArray(float* data)
  {
    fTSold->getData(fSampleFreq , data);
    fTSnew->getData(fSampleFreq , data + fSampleFreq);
  }

  /// Reset the Histogram information
  void ResetHist(const Time &time);
};
