//
//    DPUSH is the application to push the data from the NDS into the 
//    shared memory partition in frame format. This source file contains
//    both the traditional (CDS Network data server) and multicast versions.
//    The multicast version is selected with the -DMULTICNDS compilation
//    flag.
//
//    Syntax:
//
//       Dpush [NDS <node>] [partition <pname>] [lbuf <nbytes>] [nbuf <n>]
//             [log <logfile>] [time <ticks>] [-nocheck]
//
//       node    Network data server node name or number.
//       pname   Shared memory partition name
//       nbytes  Length of shared memory buffers
//       n       Number of shared memory buffers
//       logfile Log file.
//       ticks   Watchdog (statistics summary) timeout interval
//
//    Notes:
//      1) Partitions aren't currently created by the program.
//
//    Revision History:
//
//      18-Feb-1999  J. Zweizig
//          Preliminary draft version. 
//
//------- A few definitions
//
//    DEFNDSIP the default NDS server IP address.
//    DEFPART  is the default partition name that is used if the "partition" 
//             command line argument is not specified.
//    CDSDATA  receive data in CDS proprietary format.
//
#if defined(sun) && !defined(__EXTENSIONS__)
#define __EXTENSIONS__
#endif

#ifdef MULTICNDS
#define DEFNDSIP "225.0.0.1"       /* multicast address */
#else
// #define CDSDATA 1
#define DEFNDSIP "131.215.115.67"  /* "albeireo.ligo.caltech.edu" */
#endif
#define DEFPART "LIGO_Online"

//--> Include till you drop.
//    Note: unistd.h included for swab(). This didn't work in old versions
//          of linux (e.g. RH6.2), but seems to be OK with any modern version.
#include <iostream>
#include <fstream>
#include <netdb.h>
#include <netinet/in.h>
#include <stdlib.h>
#include <stdio.h>
#include <string>
#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#include "SigFlag.hh"
#include "lsmp_prod.hh"
#ifdef MULTICNDS
  #include "framexmit/framexmit.hh"
#else
  #include "DAQSocket.hh"
#endif

//--------------------------------------  Sun stuff needed to set R/T priority
#ifdef sun
  #include <sys/priocntl.h>
  #include <sys/rtpriocntl.h>
  #include <sys/mman.h>
  #include <strings.h>
#endif
#include "swapframe.hh"
#include <stdexcept>

using namespace std;

//------- Statistics class.
class Stats {
public:
  typedef unsigned long ulong_t;
  Stats();
  ~Stats();
  void Log(ostream& out);
  void Reset(void);
  void Start(ulong_t Secs);
  bool Test(void);
  void Update(ulong_t FrameID, ulong_t GPS, ostream& out);
private:
  ulong_t nSecs;
  SigFlag watchdog;
  time_t  Stamp;
  ulong_t GPSstart;
  ulong_t GPSlast;
  ulong_t RecIDstart;
  ulong_t RecIDlast;
  ulong_t FrameCount;
  ulong_t FrameLost;
  ulong_t FrameUnsent;
};

//
//    NDS interface stuff
//
#ifdef MULTICNDS
static framexmit::frameRecv* NDS = 0;
#else
static DAQSocket* NDS = 0;
#endif

//---->  Data swapper
//static SwapFrame SF;

//---->  Shared memory output device
static LSMP_PROD* OPart=0;

//---->  Termination signal flag pointer
static SigFlag term;
static bool    active;
static bool    verbose(false);
static bool    debug(false);
static bool    datacheck(true);

//---->  Watchdog statistics timer.
static Stats wstat;
unsigned long nTotalFrames(0);
unsigned long nTotalErrors(0);
unsigned long nTotalTime(0);
unsigned long lastGPS(0);

//---->  Error log control;
static bool gLog(false);
static ofstream logfi;

//----> Internal entry points.
//
//    Dpsh_Init     Set up the input and output functions
//    Dpsh_Halt     Set down.
//    Dpsh_Read     Read routine.
//    Dpsh_Check    Check a frame is consistent.
//    stamp         Generate a time stamp.
static bool Dpsh_Init(int argc, const char **argv);
static void Dpsh_Halt(void);
static void Dpsh_Read(void);
static bool Dpsh_Check(unsigned int length, const char* addr);
static string stamp(time_t now);

//
//    main function. 
//      * Set up. This includes initializing the input and output 
//        files and the exception handling.
//      * Loop over records until the ceiling falls in on us
//      * Call Dpsh_Read to process a frame
//      * When the roof caves in set-down the application
//
int main(int argc, const char **argv) {

    active = !Dpsh_Init(argc, argv);

    while (active && !term) {
        if (gLog && wstat.Test()) wstat.Log(logfi);
        Dpsh_Read();
    }
    Dpsh_Halt();
    return 0;}

//
//    Dpsh_Init - set up the internal environment for pushing data.
//       * Parse any command line parameters.
//       * Create a shared memory partition for data distribution.
//       * Make a connection to the CDS Network Data Server. Register 
//         notification of all frames.
//       * Set up the termination signal handling.
//
static bool Dpsh_Init(int argc, const char **argv) {
    const char*   FileName = (const char*)0;
    int   nbuf = 1, lbuf = 1000000, Secs=60;

#ifdef sun
    //---------------------------------  Don't run as root
    seteuid(getuid());
#endif

    //---------------------------------  Define default parameters
    const char*   ParName = getenv("LIGOSMPART");
    if (!ParName) ParName = DEFPART;
#ifdef MULTICNDS
    const char*   NDS_IP  = getenv("LIGOMCASTIP");
#else
    const char*   NDS_IP  = getenv("LIGONDSIP");
#endif
    if (!NDS_IP)  NDS_IP  = DEFNDSIP;

    //---------------------------------  Parse the arguments
    for (int i=1 ; i<argc ; i++) {
        string argi = argv[i];
        if (argi == "NDS") {
	    NDS_IP   = argv[++i];
	} else if (argi == "partition") {
	    ParName  = argv[++i];
	} else if (argi == "file") {
	    FileName = argv[++i];
	} else if (argi == "lbuf") {
	    lbuf = strtol(argv[++i], 0, 0);
	} else if (argi == "nbuf") {
	    nbuf = strtol(argv[++i], 0, 0);
	} else if (argi == "time") {
	    Secs = strtol(argv[++i], 0, 0);
	} else if (argi == "verbose") {
	    verbose = true;
	} else if (argi == "nocheck") {
	    datacheck = false;
	} else if (argi == "debug") {
	    debug = true;
	} else if (argi == "log") {
	    gLog = true;
	    logfi.open(argv[++i], ios::out);
	    logfi << stamp(0) << argv[0] << " Starting..." << endl;
	} else {
	    cerr << argv[0] << ": Invalid argument name " << argv[i] << endl;
	    cerr << "Command syntax:" << endl << endl;
	    cerr << argv[0] << " [NDS <server>] [partition <partition>] "
		 << "[file <input-file>] [debug] [nocheck] \\" << endl;
	    cerr << "       [nbuf <#buffers>] [lbuf <buffer-length>] "
		 << "[log <log-file>] [time <stat-timer>]" << endl;
	    return true;
	}
    }

    //---------------------------------  Create a Frame output
    if (FileName) {
        cerr << "File output is not supported by Dpush." << endl;
        if (gLog) logfi << stamp(0) << "File output isn't supported by Dpush."
		      << endl;
	return true;
    } else {
	OPart = new LSMP_PROD(ParName, nbuf, lbuf);
	if (!OPart->valid()) {
	    cerr << "Unable to access partition " << ParName << ": " 
		 << OPart->Error() << "." << endl;
	    if (gLog) logfi << stamp(0) << "Unable to access partition " 
			  << ParName << ": " << OPart->Error() << "." << endl;
	    return true;
	}
	OPart->lock(true);
    }

    //---------------------------------  Set up the statistics task
    if (gLog) wstat.Start(Secs);

    //---------------------------------  Set up the termination signals
    term.add(SIGINT);
    term.add(SIGTERM);
    term.add(SIGHUP);

    //---------------------------------  Open the NDS connection.
#ifdef MULTICNDS
    const char* inet = getenv("LIGONDSIP");
    if (!inet)  inet = "10.4.0.0";
    char addr[16];
    struct hostent* ptr = gethostbyname(inet);
    if (ptr) {
        inet = (const char*) ptr->h_addr_list[0];
	sprintf(addr, "%i.%i.0.0", inet[0], inet[1]);
	inet = addr;
    }
    cout << "Opening connection to frame multicaster at " << NDS_IP 
	 << " on network with " << inet << endl;
    if (gLog) logfi << stamp(0) << "Opening connection to frame multicaster: "
		  << NDS_IP << " on network with " << inet << endl;
    NDS = new framexmit::frameRecv();
    //  if (debug) NDS->setDebug(1); // not currently a method.
    if (!NDS->open(NDS_IP, inet)) {
        perror("Multicast receiver open failed.");
	return true;
    }
#else
    cout << "Opening connection to NDS at " << NDS_IP << endl;
    NDS = new DAQSocket(NDS_IP);
    NDS->setDebug(debug);
    NDS->AddChannel("all");
#ifndef CDSDATA
    NDS->RequestFrames();
    cout << "Frame data requested" << endl;
    if (gLog) logfi << stamp(0) << "Frame data requested." << endl;
#else
    NDS->RequestData();
    cout << "CDS proprietary data requested" << endl;
    if (gLog) logfi << stamp(0) << "CDS proprietary data requested" << endl;
#endif
#endif

#ifdef sun
    //---------------------------------  Establish RealTime priority
    int sveuid = geteuid();
    seteuid(0);
    if (!geteuid()) {
        pcinfo_t pci;
	pcparms_t pcp;
	rtparms_t rtp;

	if (gLog) logfi << stamp(0) << "Setting RT priority." << endl;
	strcpy(pci.pc_clname, "RT");
	int ret = priocntl(P_PID, getpid(), PC_GETCID, (char*)&pci);
	if (ret < 0) {
	    perror("Error getting CID");
	    if (gLog) logfi << stamp(0) << "Error getting CID." << endl;
	} else {
	    pcp.pc_cid = pci.pc_cid;
	    rtp.rt_pri = *(short*)&pci.pc_clinfo[0];
	    rtp.rt_tqsecs = 0;
	    rtp.rt_tqnsecs = RT_TQINF;

	    memcpy(pcp.pc_clparms, &rtp, sizeof(rtparms_t));
	    ret = priocntl(P_PID, getpid(), PC_SETPARMS, (char*)&pcp);
	    if (ret < 0) {
	        perror("Error setting params");
		if (gLog) logfi << stamp(0)<<"Error setting priority." << endl;
	    }
	}
	mlockall(MCL_CURRENT | MCL_FUTURE);
	seteuid(sveuid);
    }
#endif

    //---------------------------------  Done, quit.
    return false;}

//
//    Dpsh_Halt - set down.
//       * Release the shared memory partition.
//       * Close the connection to the CDS Network Data Server.
//       * Cancel the termination signal handling.
//
static void Dpsh_Halt(void) {

    //---------------------------------  Disable the signal
    term.zero();

    //---------------------------------  Close the NDS connection.
    if (NDS) delete NDS;

    //---------------------------------  Delete the partition
    if (OPart) delete OPart;

    //---------------------------------  Final statistics
    wstat.Log(logfi);

    //---------------------------------  Close the error log
    if (gLog) {
        gLog = false;
	logfi << stamp(0) << "Total frames read:  " << nTotalFrames << endl;
	logfi << stamp(0) << "Total frame errors: " << nTotalErrors << endl;
	logfi << stamp(0) << "Total data seconds: " << nTotalTime   << endl;
	logfi << stamp(0) << "End data GPS:       " << lastGPS      << endl;
	logfi << stamp(0) << "Dpush terminating. Error log closed." << endl;
        logfi.close();
    } else {
	cout << " Total frames read:  " << nTotalFrames << endl;
	cout << " Total frame errors: " << nTotalErrors << endl;
	cout << " Total data seconds: " << nTotalTime   << endl;
	cout << " End data GPS:       " << lastGPS      << endl;
	cout << " Dpush terminating. Error log closed." << endl;
    }
}

//
//    Dpsh_Read - read the data.
//      *  Wait for a frame to arrive.
//      *  Read in the data
//      *  Pack data into a frame
//      *  Write frame to shared memory.
//
static void Dpsh_Read(void) {
    //---------------------------------  Wait for a data message
#ifdef MULTICNDS
    // bool rc = NDS->poll();  // this just purges a buffer
    // if (!rc) return;
#else
    int rc = NDS->WaitforData(); // allow a signal before data is received.
    if (rc <= 0) return;
#endif

    //---------------------------------  Get a buffer
    char* buffer = OPart->get_buffer();
    if (!buffer) {
        active = false;
	return;
    }

    //---------------------------------  Copy the data.
    unsigned int Frame, GPS, duration;
#ifdef MULTICNDS
    int length = NDS->receive(buffer, OPart->getBufferLength(), &Frame, &GPS, 
			      &duration);

    if (length <= 0) {
        OPart->return_buffer();
	switch (length) {
	case 0:
	    cerr << "Framerecv error 0. Retry." << endl;
	    if (gLog) logfi << stamp(0) << "Framerecv error 0. Retry." << endl;
	    return;
	case -1:
	    perror("Socket assignment failed");
	    if (gLog) logfi << stamp(0) << "Socket assignment failed" << endl;
	    break;
	case -2:
	    perror("Receive buffer allocation failed");
	    if (gLog) logfi << stamp(0) << "Receive buffer allocation failed" 
			  << endl;
	    break;
	case -3:
	    perror("Unable to get packet");
	    if (gLog) logfi << stamp(0) << "Unable to get packet." 
			   << endl;
	    break;
	case -5:
	    cerr << "Invalid data array" << endl;
	    if (gLog) logfi << stamp(0) << "Invalid data array." << endl;
	    break;
	case -6:
	    cerr << "Retransmission request failed." << endl;
	    if (gLog) logfi << stamp(0) << "Retransmission request failed." 
			   << endl;
	    break;
	default:
	    cerr << "Frame length (" << -length << ") too long for buffer." 
		 << endl;
	    if (gLog) logfi << stamp(0) << "Frame length (" << -length 
			   <<") too long for buffer."<< endl;
	}
	active = false;
	return;
    } else if (debug) {
        cout << "Frame read: length/sequence/GPS/duration = " << length 
	     << "/" << Frame << "/" << GPS << "/" << duration << endl;
    }
#else
    DAQDRecHdr hdr;
#ifndef CDSDATA
    int length = NDS->GetFrame(buffer, OPart->getBufferLength(), &hdr);
#else
    int length = NDS->GetData(buffer, OPart->getBufferLength());
    hdr = *((DAQDRecHdr*)buffer);
#endif
    if (length <= 0) {
        OPart->return_buffer();
	cerr << "Receive error, rc =" << length << endl;
	if (gLog) logfi << stamp(0) << "Receive error, rc =" << length << endl;
	active = false;
	return;
    }
    Frame    = hdr.SeqNum;
    GPS      = hdr.GPS;
    duration = hdr.Secs;
#endif
    nTotalFrames++;

    //try {
    //    SF.Swap(buffer, length);
    //} catch (runtime_error& e) {
    //    cout << "Exception in SwapFrame: " << e.what() << endl;
    //	  active = false;
    //	  return;
    //}

    if (!Dpsh_Check(length, buffer) && datacheck) {
        OPart->return_buffer();
	if (gLog) {
	    logfi << stamp(0) << "Data check failed on frame " << Frame 
		 << endl;
#ifdef MULTICNDS
	    logfi << NDS->logmsg() << endl;
#endif
	} else {
	    cerr << "Data check failed on frame " << Frame << endl;
	}
	nTotalErrors++;
    } else {
        OPart->SetID(GPS);
        OPart->release(length);
	nTotalTime += duration;
	lastGPS = GPS + duration;
    }

    //----------------------------------  Check for discontinuities
    if (gLog) wstat.Update(Frame, GPS, logfi);
    return;
}

//=======================     Frame Check Code      =======================
#define MAX_ID 256
#define LSTRUCTNAME 32

inline void
ReadInc(const char*& addr, unsigned short& x, bool bs) {
    if (bs) swab(addr, &x, sizeof(x));
    else    memcpy(&x, addr, sizeof(x)); 
    addr += sizeof(x);
}

typedef long long int8_type;
struct IGWhdr {
    unsigned long len;
    short id;
    int inst;
};

static bool  DefID[MAX_ID];
static short IDEOFile;
static short IDEOFrame;
static unsigned int hdrSz;

static bool
procSH(const IGWhdr& h, const char* addr, bool bswap) {
    unsigned short lname, ID, lcomment;
    unsigned int vblen = h.len - hdrSz - 6;
    addr += hdrSz;

    ReadInc(addr, lname, bswap);
    if ((lname < 2) || (lname > vblen)) {
        if (gLog) logfi << "Inconsistent name length in SH, l=" << lname 
			<< endl;
	return true;
    }
    const char* name  = addr;
    addr += lname;

    ReadInc(addr, ID, bswap);

    ReadInc(addr, lcomment, bswap);
    if (lcomment != vblen-lname) {
        if (gLog) logfi << "Inconsistent comment length in SH, l=" << lname 
			<< endl;
	return true;
    }
    if (!ID || ID > MAX_ID) {
        if (gLog) logfi << "Illegal structure ID(" << ID << ") defined" 
			<< endl;
	return true;
    }
    DefID[ID] = true;
    if (!strncmp(name, "FrEndOf", 7)) {
        if (!strcmp(name, "FrEndOfFile"))       IDEOFile  = ID;
	else if (!strcmp(name, "FrEndOfFrame")) IDEOFrame = ID;
    }
    return false;
}

//
//  Dpsh_Check  Checks the incoming frame data structure headers for 
//              consistency.
//
//  The structure headers must meet the following requirements
//  1) The length must be greater than 8 (the length of the header) and 
//     less than the length of the Frame.
//  2) The structure type must be 1 or 2 or have been defined in a previous
//     SH structure. The Structure ID must be in the range 0 < ID < MAX_ID 
//     where MAX_ID is arbitrarily set to 256 (for now).
//  3) the structure instance numbers must increase by one with each new 
//     instance and start at either 0 or 1.
//
//  In addition the SH structures are unpacked and must be internally 
//  consistent (see procSH).
//
//  Dpsh_Check returns true if the frame passes the requirements.
//
static bool
Dpsh_Check(unsigned int length, const char* addr) {
#ifndef CDSDATA

    const char*   In   = addr;
    unsigned long left = length;
    IGWhdr h;

    //---------------------------------  Check the IGWD specifier
    if (left < 40 || strncmp("IGWD", In, 4)) {
        if (gLog) logfi << "Invalid IGWD header" << endl;
	return false;
    }
    bool bSwap = *((const short*) (In+12)) != 0x1234;
    int FrVsn = In[5];
    if (FrVsn > 4) hdrSz = 14;
    else           hdrSz = 8;
    In   += 40;
    left -= 40;

    //---------------------------------  Initialize the instance counts
    int strinst[MAX_ID];
    for (int i=0 ; i<MAX_ID ; i++) {
        strinst[i] = 0;
	DefID[i] = false;
    }
    DefID[1] = true;
    DefID[2] = true;

    //---------------------------------  Loop over structures
    while (1) {
        if (left < sizeof(h)) {
	    if (gLog) logfi << "Unexpected EOD." << endl;
	    return false;
	}

	if (FrVsn <=4) {
	    struct {
	      int l;
	      short s;
	      short i;
	    } h4;

	    if (bSwap) {
	        swab(In, (char*)&h4, sizeof(h4));
		short* sptr = reinterpret_cast<short*>(&h4);
		short  ts = sptr[0];
		sptr[0]  = sptr[1];
		sptr[1]  = ts;
	    } else {
	        memcpy(&h4, In, sizeof(h4));
	    }
	    h.len  = h4.l;
	    h.id   = h4.s;
	    h.inst = h4.i;

	} else {
	    union {
	      int8_type l;
	      short     s[4];
	    } hlen;

	    union {
	      int    i;
	      short  s[2];
	    } hinst;

	    short hid;

	    if (bSwap) {
	        short hw[7];
	        swab(In, (char*)&hw, sizeof(hw));
		hlen.s[0]  = hw[3];
		hlen.s[1]  = hw[2];
		hlen.s[2]  = hw[1];
		hlen.s[3]  = hw[0];
		hid        = hw[4];
	        hinst.s[0] = hw[6];
	        hinst.s[1] = hw[5];		
	    } else {
	        memcpy(hlen.s,  In   , sizeof(hlen.l));
	        memcpy(&hid,    In+ 8, sizeof(hid));
	        memcpy(hinst.s, In+10, sizeof(hinst.i));
	    }

	    h.len  = hlen.l;
	    h.id   = hid;
	    h.inst = hinst.i;
	}

	if (h.len < hdrSz || h.len > left) {
	    if (gLog) logfi << "Illegal structure length " << h.len 
			   << " at offset " << length-left << endl;
	    return false;
	}

	if (h.id >= MAX_ID || !DefID[h.id]) {
	    if (gLog) logfi << "Illegal or undefined structure ID " << h.id 
			  << " at offset " << length-left << endl;
	    return false;
	}

	if (h.inst != strinst[h.id]++) {
	    if (h.inst == 1 && strinst[h.id] == 1) {
	        strinst[h.id]++;
	    } else {
	        if (gLog) logfi << "Invalid structure instance " << h.inst 
			      << " of ID " << h.id << ", expected " 
			      << strinst[h.id]-1 << " at offset " 
			      << length-left << endl;
		return false;
	    }
	}

	if ((h.id == 1) && procSH(h, In, bSwap)) return false;

	In   += h.len;
	left -= h.len;
#ifndef FRAMELIB
	if (h.id == IDEOFrame) {
	    if (FrVsn >= 4) {
	        if (FrVsn>=6) for (int i=3 ; i<MAX_ID ; i++) strinst[i] = 0;
		else          for (int i=1 ; i<MAX_ID ; i++) strinst[i] = 0;
	    }
	}
#endif
	if (h.id == IDEOFile) break;
    }
    if (left) {
        if (gLog) logfi << "Data left after end-of-frame." << endl;
	return false;
    }
#endif  // !def(CDSDATA)
    return true;
}

static string stamp(time_t now) {
    if (now == 0) now = time(0);
    string datel(::ctime(&now)+4, 15);
    datel += " ";
    return datel;
}

//--------------------------------------  Statistics class methods
Stats::Stats(void) 
  : GPSlast(0), RecIDlast(0)
{}

Stats::~Stats() {
}

bool 
Stats::Test() {
  return watchdog;
}

void 
Stats::Log(ostream& out) {
    time_t now = time(0);
    float Rate;
    if (now == Stamp) Rate = 1.0;
    else              Rate = float(GPSlast - GPSstart) / float(now-Stamp);
    if (FrameLost || FrameUnsent || Rate > 1.02 || Rate < 0.98) {
        out << stamp(now) << " Statistics - Frames read: " 
	    << FrameCount << " Lost: " << FrameLost << " Unsent: " 
	    << FrameUnsent << " Rate: " << Rate << endl;
    }
    if (debug) {
        out << "Lock delay count = " << OPart->getLockCount() << endl;
	OPart->clrLockCount();
    }
    Reset();
    watchdog.clear();
}

void 
Stats::Reset(void) {
    GPSstart   = GPSlast;
    RecIDstart = RecIDlast;
    FrameCount = 0;
    FrameLost  = 0;
    FrameUnsent = 0;
    ::time(&Stamp);
}

void 
Stats::Start(ulong_t Secs) {
    struct itimerval ticks;

    ticks.it_value.tv_sec  = Secs;
    ticks.it_value.tv_usec  = 0;
    ticks.it_interval.tv_sec  = Secs;
    ticks.it_interval.tv_usec  = 0;
    nSecs = Secs;
    setitimer(ITIMER_REAL, &ticks, 0);
    watchdog.setMode(SigFlag::kBlock);
    watchdog.add(SIGALRM);
    time(&Stamp);
}

void 
Stats::Update(ulong_t FrameID, ulong_t GPS, ostream& log) {
    if (!GPSlast) {
        log << stamp(0) << "First frame read, GPS: " << GPS 
	    << " Frame: " << FrameID << endl;
        GPSlast   = GPS;
	RecIDlast = FrameID;
	Reset();
    } else if (FrameID == 1) {
        log << stamp(0) << "Frame broadcaster restarted, first GPS= " << GPS
	    << " previous= " << GPSlast << endl;
    } else if (FrameID <= RecIDlast) {
        log << stamp(0) << "Frame received out of order, Frame=" << FrameID
	    << " expected " << RecIDlast+1 << endl;
    } else if (GPS <= GPSlast) {
        log << stamp(0) << "Frame received out of order, GPS=" << GPS 
	    << " expected " << GPSlast+1 << endl;
    } else {
        if (verbose && GPS != GPSlast+1) {
	    log << "Missing frame " << GPSlast+1;
	    if (GPS-GPSlast>2) log << "-" << GPS-1;
	    log << endl;
	}
	ulong_t lost = FrameID - RecIDlast - 1;
        FrameLost   += lost;
	FrameUnsent += GPS - GPSlast - 1 - lost;
    }
    FrameCount++;
    GPSlast   = GPS;
    RecIDlast = FrameID;
}
