/* -*- mode: c++; c-basic-offset: 4; -*- */
#include "LdasDBWriter.hh"
#include "Segment.hh"
#include "TrigBase.hh"
#include "TrigProc.hh"
#include "TrigTable.hh"
#include "xsil/Xwriter.hh"
#include "xsil/ligolw.hh"
#include "lmsg/ErrorList.hh"
#include <fstream>
#include <sstream>
#include <algorithm>

using namespace std;

namespace trig {

  //====================================  Ldas writer constructor
  LdasDBWriter::LdasDBWriter(void) 
    : mNPrcWrt(0)
  {
  }

  //====================================  Ldas writer destructor
  LdasDBWriter::~LdasDBWriter(void) {
  }

  //====================================  Add a trigger to the list
  lmsg::error_type 
  LdasDBWriter::addTrigger(const TrigBase& t, const TrigProc& p) {
      proc_iter pit = insert_proc(p);
      if (!mTrgList.empty() && mTrgList.back() == t) {
	  cerr << "Trigger repeated. ID: " << t.getID() 
	       << " SubID: " << t.getSubID()  << endl;
	  return lmsg::Invalid;
      }
      mTrgList.push_back(t);
      mTrgList.back().setProcess(pit->getProcessID());
      return lmsg::OK;
  }

  //====================================  Add a segment to the list
  lmsg::error_type 
  LdasDBWriter::addSegment(const Segment& s,  const TrigProc& p) {
      if (s.getActivity() <= 0) return lmsg::OK;
      proc_iter pit = insert_proc(p);
      if (!mSegList.empty() && mSegList.back() == s) {
	  cerr << "Segment repeated. Group: " << s.getGroup() << endl;
	  return lmsg::Invalid;
      }
      mSegList.push_back(s);
      mSegList.back().setProcess(pit->getProcessID());
      return lmsg::OK;
  }

  //====================================  Clear temporary tables
  void 
  LdasDBWriter::clear(void) {
      // mPrcList.clear();
      mNPrcWrt = mPrcList.size();
      mSegList.clear();
      mTrgList.clear();
  }

  //======================================  Data size methods
  bool
  LdasDBWriter::empty(void) const {
    return mTrgList.empty() && mSegList.empty();
  }

  int
  LdasDBWriter::getNSegs(void) const {
    return mSegList.size();
  }

  int
  LdasDBWriter::getNTrigs(void) const {
    return mTrgList.size();
  }

  //====================================  Add a segment to the list
  lmsg::error_type 
  LdasDBWriter::setProcess(const TrigProc& p) {
      refProcess()  = p;
      Time idTime(Now());
      idTime.setN(mPrcList.size()+1); // replace ns field with unique ID
      char line[128];
      TimeStr(idTime, line, "%Y%02m%02d%02H%02N%02S%06n010000");
      refProcess().setProcessID(line);
      proc_iter pit = insert_proc(refProcess());
      return lmsg::OK;
  }

  //====================================  Write out tables as xml.
  lmsg::error_type 
  LdasDBWriter::write(const std::string& file) const {
      if (mTrgList.empty() && mSegList.empty()) return lmsg::OK;
      ofstream out(file.c_str());
      if (!out.is_open()) {
	  cerr << "LdasDBWriter is unable to open file: " << file << endl;
	  return lmsg::Failure;
      }
      xsil::Xwriter xw(out);
      xw.setDocType("SYSTEM \"http://gateway/doc/ligolwAPI/html/ligolw_dtd.txt\"");
      xsil::ligolw lwfile("ligo:ldas:file");
 
      //--------------------------------  Fill in the (new) processes
      if (int(mPrcList.size()) > mNPrcWrt) {
	  const_proc_iter pStart = mPrcList.begin();
	  for (int i=0; i<mNPrcWrt; ++i) ++pStart;

	  ProcTable* pTab = new ProcTable(false);
	  for (const_proc_iter i=pStart; i!=mPrcList.end(); ++i) {
	      pTab->addRow(*i);
	      //mNPrcWrt++;
	  }
	  lwfile.addObject(pTab);
      }
 
      //--------------------------------  Fill in the (new) processes
      if (!mSegList.empty()) {
	  SegTable* sTab = new SegTable(false);
	  for (const_seg_iter i=mSegList.begin(); i!=mSegList.end(); ++i) {
	      sTab->addRow(*i);
	  }
	  lwfile.addObject(sTab);
      }
 
      //--------------------------------  Fill in the (new) processes
      if (!mTrgList.empty()) {
	TrigTable* tTab = new TrigTable(false);
	  for (const_trig_iter i=mTrgList.begin(); i!=mTrgList.end(); ++i) {
	      tTab->addRow(*i);
	  }
	  lwfile.addObject(tTab);
      }
      lwfile.Spew(xw);
      if (out.fail()) {
	  cerr << "LdasDBWriter: Writing triggers/segments to file: " 
	       << file << " failed." << endl;
	  return lmsg::Failure;
      }
      return lmsg::OK;
  }

  //------------------------------------  Add/find process in list
  LdasDBWriter::proc_iter 
  LdasDBWriter::insert_proc(const TrigProc& p) {
      proc_iter pit = find(mPrcList.begin(), mPrcList.end(), p);
      if (pit == mPrcList.end()) {
	  int pid = mPrcList.size();
	  mPrcList.push_back(p);
	  pit = mPrcList.end(); --pit;
	  if ( ! *(pit->getProcessID())) {
	      ostringstream id;
	      id << "process:process_id:" << pid;
	      pit->setProcessID(id.str());
	  }
	  if (getDebug()) {
	      cerr << "LdasDBWriter: Added process: " << pit->getName()
		   << " pID: " << pit->getProcessID() << endl;
	  }
      } else if (getDebug() > 1) {
	  cerr << "LdasDBWriter: Found process: " << pit->getName()
	       << " pID: " << pit->getProcessID() << endl;
      }
      return pit;
  }

} // namespace trig
