/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef CHANNEL_HH
#define CHANNEL_HH

#include "Time.hh"
#include "Interval.hh"
#include <string>

#ifndef __CINT__
#include "framecpp/Types.hh"
#else
namespace FrameCPP {
  class FrameH;
  class FrAdcData;
  class FrProcData;
  class FrSimData;
  class FrVect;
}
#endif  // !def(__CINT__)

class TSeries;
class FSeries;
class FrVectRef;

/**  Channel holds the information about the Time Series being accumulated 
  *  by Dacc.
  *  @memo Channel list entry for Dacc.
  *  @author John Zweizig
  *  @version 1.1; Last modified: August 12, 2008
  */
class Channel 
{
public:
  typedef unsigned int uint_type;
  typedef unsigned long size_type;
  enum ChanType {
    kUnknown,
    kRaw,
    kProcessed,
    kSimulated,
    kFSeries
  };

  //------------------------------------  Constructor/Destructor.
public:
  /**  Default Constructor.
    */
  Channel(void);

  /**  @memo Construct a channel object.
    *  Construct an object given a channel name, a decimation fator, 
    *  an output pointer address and a channel type.
    */
  Channel(const char* name, TSeries** ctlvar, uint_type decim=0, 
	  ChanType t=kUnknown);

  /**  @memo Construct a channel object.
    *  Construct an object given a channel name, a decimation fator, 
    *  an output pointer address and a channel type.
    */
  Channel(const char* name, FSeries** ctlvar);

  /**  @memo Copy constructor.
    *  Construct a channel object identical to an already existing one.
    */
  Channel(const Channel& data);

  /**  @memo Assignment operator.
    *  Copy a channel object to an existing object.
    */
  const Channel& operator=(const Channel& data);

  /**  Destroy a channel entry.
    */
  ~Channel();

  //------------------------------------  Accessors
  /**  @memo Compare the name.
    *  EqName() returns true if the channel name is the same as the argument.
    */
  bool EqName(const char *name) const;

  /**  @memo Compare the name and type.
    *  Equal returns true if the channel name is the same as the argument 
    *  and the channel type is the same, or the requested type is unknown.
    */
  bool Equal(const char *name, ChanType t=kUnknown) const;

  /**  @memo Get the debug flag.
    *  getDebug() returns the debug flag. When the debug flag is set, 
    *  Channel methods print messages to cerr any time an error occurs.
    */
  bool getDebug(void) const;

  /**  @memo Get the decimation factor.
    *  getDecim() returns the decimation factor.
    */
  uint_type getDecim(void) const;

  /**  @memo Get the end time.
    *  getLast() returns the time of the last sample added to the channel 
    *  time series.
    */
  Time getLast(void) const;

  /**  @memo Get the channel name.
    *  getName() returns a pointer to the channel name string.
    */
  const char* getName(void) const;

  /**  @memo Get the channel type.
    *  The channel type is returned. The type can be #kUnknown# if the 
    *  type has not been specified and no data from this channel has been 
    *  read, #kRaw# if the channel is a raw (Adc) data channel or 
    *  #kProcessed# if the data are in a process data channel.
    */
  ChanType getType(void) const;


  //------------------------------------  Mutators
  /**  Set channel accumulation mode.
    *  setAccumulate() sets/clears the accumulate mode for the channel.
    *  If accumulate mode is set, the data will be appended to the TSeries. 
    *  By default the TSeries data are reset with each call to FillData().
    */
  void setAccumulate(bool Acc=true) {mAccumulate = Acc;}

  /**  Set/clear the debug flag.
    *  The debug flag is set as specified.
    */
  void setDebug(bool yorn) {mDebug = yorn;}

  //------------------------------------  Crunchers
  /**  Allocate an output TSeries with a specified length.
    *  The user TSeries data vector is expanded to accommodate data over
    *  a specified time interval.
    */
  void allocate(Interval Stride);

  /**  Data from the specified frame are (optionally) decimated and copied 
    *  into the data vector(s).
    *  @memo Fill time series from a frame in memory.
    *  @return 
    *  \begin{tabular}{|r|l|} \\
          0 & Successful completion \\
	 -1 & Frame start not contiguous to previous data \\
	 -2 & Sample rate isn't compatible with previous data. \\
	 -3 & Requested data not found in current frame \\
	 -5 & Frame data are not self-consistent. \\
	 -6 & TSeries is not allocated. \\
	 -7 & Unsupported data type \\
    *  \end{tabular}
    *  @param frame Pointer to frame containing data to be read.
    *  @param off   Time offset to tstart of data.
    *  @param dT    Time interval to be read in.
    */
  int  FillSeries(const FrameCPP::FrameH* frame, Interval off, Interval dT);

  /**  Data from the specified adc block are (optionally) decimated and 
    *  copied into the data vector(s).
    *  @memo Fill time series from an adcData object.
    *  @return 
    *  \begin{tabular}{|r|l|} \\
          0 & Successful completion \\
	 -1 & Frame start not contiguous to previous data \\
	 -2 & Sample rate isn't compatible with previous data. \\
	 -3 & Requested data not found in current frame \\
	 -5 & Frame data are not self-consistent. \\
	 -6 & TSeries is not allocated. \\
	 -7 & Unsupported data type \\
    *  \end{tabular}
    *  @param adc   Pointer to FrAdcData containing data.
    *  @param t0    Start time for data unpacking.
    *  @param off   Time offset to tstart of data.
    *  @param dT    Time interval to be read in.
    */
  int  FillSeries(const FrameCPP::FrAdcData* adc, const Time& t0, 
		  Interval off, Interval dT);

  /**  Data from the specified FrProcData block are (optionally) decimated 
    *  and copied into the data vector(s).
    *  @memo Fill time series from an FrProcData object.
    *  @return 
    *  \begin{tabular}{|r|l|} \\
          0 & Successful completion \\
	 -1 & Frame start not contiguous to previous data \\
	 -2 & Sample rate isn't compatible with previous data. \\
	 -3 & Requested data not found in current frame \\
	 -5 & Frame data are not self-consistent. \\
	 -6 & TSeries is not allocated. \\
	 -7 & Unsupported data type \\
    *  \end{tabular}
    *  @param proc  Pointer to FrProcData containing data.
    *  @param t0    Start time for data unpacking.
    *  @param off   Time offset to tstart of data.
    *  @param dT    Time interval to be read in.
    */
  int  FillSeries(const FrameCPP::FrProcData* proc, const Time& t0, 
		  Interval off, Interval dT);

  /**  Data from the specified FrSimData block are (optionally) decimated 
    *  and copied into the data vector(s).
    *  @memo Fill time series from an FrSimData object.
    *  @return 
    *  \begin{tabular}{|r|l|} \\
          0 & Successful completion \\
	 -1 & Frame start not contiguous to previous data \\
	 -2 & Sample rate isn't compatible with previous data. \\
	 -3 & Requested data not found in current frame \\
	 -5 & Frame data are not self-consistent. \\
	 -6 & TSeries is not allocated. \\
	 -7 & Unsupported data type \\
    *  \end{tabular}
    *  @param proc  Pointer to FrSimData containing data.
    *  @param t0    Start time for data unpacking.
    *  @param off   Time offset to tstart of data.
    *  @param dT    Time interval to be read in.
    */
  int  FillSeries(const FrameCPP::FrSimData* proc, const Time& t0, 
		  Interval off, Interval dT);

  /**  A pointer to the filled time series is returned.
    *  @memo Reference the TSeries.
    */
  const TSeries* refSeries(void) const {return (mUserPtr)? *mUserPtr : 0;}
  TSeries* refSeries(void) {return (mUserPtr)? *mUserPtr : 0;}

  /**  Return a pointer to the frequency series filled by the last 
    *  call to FillSeries.
    *  @memo Reference the FSeries.
    */
  const FSeries* refFSeries(void) const {return (mFUserPtr)? *mFUserPtr : 0;}
  FSeries* refFSeries(void) {return (mFUserPtr)? *mFUserPtr : 0;}

  /**  Sufficient space is reserved in the target time series to accomodate 
    *  the specified interval. No action is taken if the sample rate is 
    *  unknown (zero) or if the target series is a frequency series.
    *  @memo resize the target series to accomodate the specified interval.
    *  @param dT Time interval for which space is to be reserved.
    */
  void reserve(Interval dT);

private:                                // Private functions.

  /**  Data from the specified FrVect block are (optionally) decimated 
    *  and copied into the data vector(s).
    *  @memo Fill time series from an FrVect object.
    *  @return 
    *  \begin{tabular}{|r|l|} \\
          0 & Successful completion \\
	 -1 & Frame start not contiguous to previous data \\
	 -2 & Sample rate isn't compatible with previous data. \\
	 -3 & Requested data not found in current frame \\
	 -5 & Frame data are not self-consistent. \\
	 -6 & TSeries is not allocated. \\
	 -7 & Unsupported data type \\
    *  \end{tabular}
    *  @param vec Pointer to FrVect containing data.
    *  @param t0  Start time for data unpacking.
    *  @param off Time offset to tstart of data.
    *  @param dT  Time interval to be read in.
    *  @param fix Fix the data for readout errors.
    *  \begin{tabular}{|r|l|} \\
          0 & No transformation is necessary \\
	  1 & Swap words to fix S2 data error. \\
    *  \end{tabular}
    *  
    */
  int  FillSeries(const FrVectRef& vec, const Time& t0, 
		  Interval off, Interval dT, int fix=0);

  /**  Data from the specified FrVect block are copied into the data vector.
    *  @memo Fill frequency series from an FrVect object.
    *  @return 
    *  \begin{tabular}{|r|l|} \\
          0 & Successful completion \\
	 -3 & Requested data not found in current frame \\
	 -5 & Frame data are not self-consistent. \\
	 -6 & FSeries is not allocated. \\
	 -7 & Unsupported data type \\
    *  \end{tabular}
    *  @param vec Pointer to FrVect containing data.
    *  @param t0  Start time for data unpacking.
    *  @param dT  Time interval to be read in.
    */
  int  FillFSeries(const FrVectRef& vec, const Time& t0, Interval dT);

  /**  Delete a locally managed series.
    *  @memo Delete a locally managed series.
    */
  void deleteSeries(void);

  /**  Compare the specified interval to the current sample interval.
    *  set mSample if it isn't set and return an error indicator.
    *  @memo Set the sample interval.
    */
  int setSample(const FrVectRef& vec);

  /**  Compare the specified interval to the current sample interval.
    *  set mSample if it isn't set and return an erro indicator.
    *  @memo Set the sample interval.
    */
  int setSample(Interval dT);

private:                                // Private data members.
  /**  Channel Name.
    *  The name of the channel to be processed.
    */
  std::string mName;

  /**  Decimation Factor.
    *  Decim8 contains the number of raw samples that must be averaged 
    *  together for a single TSeries entry.
    */
  uint_type mDecim8;

  /**  Accumulation mode flag.
    *  If mAccumulate is true, each call to FillSeries results in the 
    *  new data being appended to the those data that are already in the
    *  target TSeries. If it is false, the new data read by fillSeries 
    *  overwrite the previous contents of the TSeries.
    */
  bool mAccumulate;

  /**  Debug flag.
    *  If set, copious error and status messages are printed to cerr.
    */
  bool mDebug;

  /**  Raw Sample time.
    *  mSample contains sample time (1/<sample-frequency>) of the channel
    *  in the input frame data.
    */
  Interval mSample;

  /**  Last sample time.
    *  The absolute (GPS) time of the end of the time bin containing the 
    *  last sample processed.
    */
  Time mLast;

  /**  @memo Number of samples added to mAccVal.
    *  The number of samples that have been added to mAccVal (in the range 
    *  0 to mDecim8-1)
    */
  uint_type mNSample;

  /**  @memo Current accumulator.
    *  The sum of the last mNSample digitizations.
    */
  double mAccVal;

  /**  @memo User TSeries pointer address.
    *  Address of the users pointer to the Time series to be filled.
    */
  TSeries** mUserPtr;

  /**  @memo User FSeries pointer address.
    *  Address of the users pointer to the Time series to be filled.
    */
  FSeries** mFUserPtr;

    /**  @memo Local pointer.
      *  Local pointer used if the user doesn't specify one.
      */
    union {
	FSeries* fsp;
	TSeries* tsp;
    } myPtr;

    /**  @memo Channel type
      *  Flag indicating type of channel (Raw or processed).
      */
    ChanType  mType;
};

inline bool 
Channel::EqName(const char *name) const {
    return (mName == name);
}

inline bool 
Channel::Equal(const char *name, ChanType t) const {
    return (mName == name && (t == kUnknown || t == mType));
}

inline bool 
Channel::getDebug(void) const {
    return mDebug;
}

inline Channel::uint_type 
Channel::getDecim(void) const {
    return mDecim8;
}

inline Time 
Channel::getLast(void) const {
    return mLast;
}

inline const char* 
Channel::getName(void) const {
    return mName.c_str();
}

inline Channel::ChanType
Channel::getType(void) const {
    return mType;
}

#endif // CHANNEL_HH
