/* -*- mode: c++; c-basic-offset: 4; -*- */
///////////////////////////////////////////////////////////////////////////////
//                                                                           //
//                                LPEFilter.cc                               //
//                                                                           //
//               Linear predictive error filter implementation               //
//                                                                           //
///////////////////////////////////////////////////////////////////////////////

#include "LPEFilter.hh"
#include "DVecType.hh"
#include "lcl_array.hh"
#include <iostream>

//--------------------------------------  Default constructor.
LPEFilter::LPEFilter(int filterLength, int trainPeriod, int trainLength)
  : FIRdft(), trainPeriod(trainPeriod), trainLength(trainLength),
    trainTime(0)
{
  setLength(filterLength);
}

//--------------------------------------  Copy constructor.
LPEFilter::LPEFilter(const LPEFilter& lpef)
  : FIRdft(lpef), trainPeriod(lpef.trainPeriod),
    trainLength(lpef.trainLength), trained(false), trainTime(0)
{
  reset();
}

//--------------------------------------  Destructor.
LPEFilter::~LPEFilter()
{
}

//======================================  Clone
LPEFilter*
LPEFilter::clone(void) const {
  return new LPEFilter(*this);
}

//--------------------------------------  Assignment operator
LPEFilter& LPEFilter::operator=(const LPEFilter& lpef)
{

  if (this != &lpef) {
    this->FIRdft::operator=(lpef);
    trainPeriod = lpef.trainPeriod;
    trainLength = lpef.trainLength;
    trainTime = Time(0);
    reset();
  }

  return *this;

}

//--------------------------------------  Apply filter to TSeries
TSeries& LPEFilter::apply(const TSeries& in, TSeries& out)
{

    if (!trainTime) {
	trained = !train(in);
    } else if (trainPeriod != 0 && in.getStartTime() >= trainTime+trainPeriod) {
	trained = !train(in);
    } else {
	trained = false;
    }

    FIRdft::apply(in, out);

    return out;
}

//--------------------------------------  Determine filter coefficients
int LPEFilter::train(const TSeries& data)
{
    int M = getLength();
    if (!M) {
	std::cerr << "LPEFilter: filter length not defined" << std::endl;
	return -1;
    }

    if (!trainLength) {
	std::cerr << "LPEFilter: training length not defined" << std::endl;
	return -1;
    }

    if (trainLength < M) {
	std::cerr << "LPEFilter: insufficient training length" << std::endl;
	return -1;
    }

    if (int(data.getNSample()) < trainLength) {
	std::cerr << "LPEFilter: insufficient training data" << std::endl;
	return -1;
    }

    lcl_array<double> x(trainLength);
    data.getData(trainLength, x);

    lcl_array<double> r(M);
    autocorr(x, r);

    lcl_array<double> c(M);
    levinson(r, c);

    setCoefs(c);

    setRate(1.0 / data.getTStep());

    trainTime = data.getStartTime();

    return 0;
}

//--------------------------------------  Compute autocorrelation vector
void LPEFilter::autocorr(const double* x, double* r) const
{

    int M = getLength();
    for (int lag = 0; lag < M; ++lag) {
	register double sum = 0.0;
	for (int sample = lag; sample < trainLength; ++sample) {
	    sum += x[sample] * x[sample - lag];
	}
	r[lag] = sum / double(trainLength - lag);
    }

}

//--------------------------------------  Solve symmetric Yule-Walker problem
void LPEFilter::levinson(const double* r, double* c) const
{

    int M = getLength() - 1;

    c[1] = - r[1] / r[0];

    for (int m = 1; m < M; ++m) {

	double num = r[m + 1];
	double den = r[0];

	for (int n = 1; n <= m; ++n) {
	    num += c[n] * r[m + 1 - n];
	    den += c[n] * r[n];
	}

	c[m + 1] = - num / den;

	int movr2 = (m+1) >> 1;
	for (int n = 1; n <= movr2; ++n) {
	    double tmp = c[n] + c[m + 1] * c[m + 1 - n];
	    c[m + 1 - n] += c[m + 1] * c[n];
	    c[n] = tmp;
	}

    }

    c[0] = 1;

}

//--------------------------------------  Set Train Period
void LPEFilter::setTrainPeriod(int period)
{
  trainPeriod = period;
}

//--------------------------------------  Set Train Length
void LPEFilter::setTrainLength(int length)
{
  trainLength = length;
}

//--------------------------------------  Set History
void LPEFilter::setHistory(const TSeries& in)
{

  reset();

  if (getRate() == 0.0) {
    setRate(1.0 / in.getTStep());
  }

  FIRdft::setHistory(in);

}

//--------------------------------------  Get Train Period
int LPEFilter::getTrainPeriod(void) const
{
  return trainPeriod;
}

//--------------------------------------  Get Train Length
int LPEFilter::getTrainLength(void) const
{
  return trainLength;
}

//--------------------------------------  Get Train Time
Time LPEFilter::getTrainTime(void) const
{
  return trainTime;
}

//--------------------------------------  Get Transient Time
Interval LPEFilter::getTransientTime(void) const
{
  return getLength() / getRate();
}
